# This Python script automatically generates the module
# diagnostics_ascii. It is automatically invoked by the Makefile;
# thus, after editing it a simple make will trigger the generation of
# the file diagnostics_ascii.f90.
#
# This could make good use of jinja templating, but we do it in a more
# manual approach currently in order to avoid external dependencies

# This is a list of all the ascii output files
ascii_files = [
  'out',
  'phase',
  'heat',
  'heat2',
  'vres',
  'lpc',
  'vres2',
  'cres',
  'parity',
  'jext',
  'eigenfunc'
]


header = """
! DO NOT EDIT THIS FILE
! This file has been automatically generated using generate_diagnostics_ascii.py

!> A module for managing text-based output files
!> for the new diagnostics module
"""

module_string = """
{header}
module diagnostics_ascii

{type}

contains

{init}

{finish}

{flush}

end module diagnostics_ascii
"""

def generate_type(files):
    type_string = """
!> Holds integers corresponding to the
!> ascii output units
type diagnostics_ascii_type
{contents}
end type diagnostics_ascii_type
    """

    unit_declaration = "  integer :: {name}"
    flag_declaration = "  logical :: write_to_{name} = .false."

    unit_contents = "\n".join([unit_declaration.format(name = name) for name in files])
    flag_contents = "\n".join([flag_declaration.format(name = name) for name in files])

    contents = "\n".join([unit_contents, flag_contents])

    return type_string.format(contents = contents).strip()

def generate_init(files):
    init_string = """
!> Opens all enabled ascii output files and records
!> file units.
subroutine init_diagnostics_ascii(ascii_files)
  use file_utils, only: open_output_file
  type(diagnostics_ascii_type), intent(in out) :: ascii_files

{contents}

end subroutine init_diagnostics_ascii
"""
    open_line = "  if (ascii_files%write_to_{name}) call open_output_file(ascii_files%{name}, '.{name}')"

    contents = "\n".join([open_line.format(name = name) for name in files])

    return init_string.format(contents = contents).strip()

def generate_finish(files):
    finish_string = """
!> Closes all enabled ascii output files as given
!> by recorded file units.
subroutine finish_diagnostics_ascii(ascii_files)
  use file_utils, only: close_output_file
  type(diagnostics_ascii_type), intent(in out) :: ascii_files

{contents}

end subroutine finish_diagnostics_ascii
"""
    close_line = "  if (ascii_files%write_to_{name}) call close_output_file(ascii_files%{name})"

    contents = "\n".join([close_line.format(name = name) for name in files])

    return finish_string.format(contents = contents).strip()


def generate_flush(files):
    flush_string = """
!> Flushes all enabled ascii output files as given
!> by recorded file units.
subroutine flush_output_files(ascii_files)
  use file_utils, only: flush_output_file
  type(diagnostics_ascii_type), intent(in out) :: ascii_files

{contents}

end subroutine flush_output_files
"""
    flush_line = "  if (ascii_files%write_to_{name}) call flush_output_file(ascii_files%{name})"

    contents = "\n".join([flush_line.format(name = name) for name in files])

    return flush_string.format(contents = contents).strip()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description='Generate new diagnostics code for dealing with ascii files.')
    parser.add_argument("output")
    args = parser.parse_args()
    with open(args.output, 'w') as output:
        output.write(module_string.format(
            header = header.strip(),
            type = generate_type(ascii_files),
            init = generate_init(ascii_files),
            finish = generate_finish(ascii_files),
            flush = generate_flush(ascii_files)
        ))
