# This Python script automatically generates the module
# diagnostics_create_and_write. It is automatically invoked by the Makefile;
# thus, after editing it a simple make will trigger the generation of
# the file diagnostics_create_and_write.f90.
#
# This could make good use of jinja templating, but we do it in a more
# manual approach currently in order to avoid external dependencies

header = """
! DO NOT EDIT THIS FILE
! This file has been automatically generated using generate_diagnostics_create_and_write.py

!> A module which contains two high level interfaces for writing
!> to the netcdf file, one for variables which are local to all
!> processors, and one for variables which may be distributed in
!> the same manner as the fields.
"""

module_string = """
{header}
module diagnostics_create_and_write
  use iso_fortran_env, only: real32, real64

  implicit none

  private

  !> Create and/or write the given variable depending on the values
  !> of the flags gnostics%create and gnostics%wryte
  public :: create_and_write_variable

  !> Create and/or write the given variable depending on the values
  !> of the flags gnostics%create and gnostics%wryte
  !> This version will not attempt to read the variable during a
  !> replay, useful for constants such as input parameters,
  !> normalisations
  public :: create_and_write_variable_noread

  !> These are a set of subroutines for writing variables which
  !> have the dimensions kx and ky (for example, fields, moments or
  !> fluxes) which may be distributed, ie. different kx,ky combinations
  !> may be on different processors. The locality is determined through
  !> the function field_k_local.
  public :: create_and_write_distributed_fieldlike_variable

{create_and_write_variable_interface}

{create_and_write_variable_noread_interface}

{create_and_write_distributed_fieldlike_variable_interface}

contains

{create_and_write_variable_implementations}

{create_and_write_variable_noread_implementations}

{create_and_write_distributed_fieldlike_variable_implementations}

end module diagnostics_create_and_write
"""

# Given a data type and dimension count determine the unique identifier
def get_basic_id(data_type, dimensions):

    return data_type.replace("(kind=","_").replace(")","")+"_"+str(dimensions)

# Used to generate the create_and_write_variable and the no-read variant
# as these routines are near identical, only practically differing in
# the intent on the data
def generate_create_and_write_variable(data_type, dimensions, is_no_read = False):
    routine_string = """
subroutine create_and_write_variable_{id}(gnostics, variable_type, variable_name, dimension_list, variable_description, variable_units, val)
  use simpledataio, only: create_variable
  use simpledataio_write, only: write_variable
  use simpledataio_read, only: read_variable
  use diagnostics_config, only: diagnostics_type
  type(diagnostics_type), intent(in) :: gnostics
  integer, intent(in) :: variable_type
  character(*), intent(in) :: variable_name
  character(*), intent(in) :: dimension_list
  character(*), intent(in) :: variable_description
  character(*), intent(in) :: variable_units
  {data_type},{dimensionality} intent({intent}) :: val

  if (gnostics%create) then
    call create_variable(gnostics%sfile, variable_type, variable_name, trim(dimension_list), variable_description, variable_units)
  end if
  {read_line}
  if (gnostics%wryte) call write_variable(gnostics%sfile, variable_name, val)

end subroutine create_and_write_variable_{id}
"""
    routine_id = get_basic_id(data_type, dimensions)

    if is_no_read:
        read_line = ""
        intent = "in"
        routine_id += "_nr"
    else:
        read_line = "if (gnostics%reed) call read_variable(gnostics%sfile, variable_name, val)"
        intent = "in out"

    if dimensions == 0 :
        dimensionality = ""
    else:
        dimensionality = " dimension("+(":, "*(dimensions-1))+":),"

    return routine_string.format(
        id = routine_id,
        dimensionality = dimensionality,
        data_type = data_type,
        intent = intent,
        read_line = read_line
    ).strip()



# Used to generate the create_and_write_distributed_fieldlike_variable routines
def generate_create_and_write_distributed_fieldlike_variable(data_type, dimensions):
    routine_string = """
subroutine create_and_write_dstrb_fieldlike_variable_{id}(gnostics, variable_type, variable_name, dimension_list, &
  variable_description, variable_units, val)
  use simpledataio, only: create_variable, set_start, set_count
  use simpledataio, only: set_independent, set_collective
  use simpledataio_write, only: write_variable, write_variable_with_offset
  use simpledataio_read, only: read_variable, read_variable_with_offset
  use diagnostics_config, only: diagnostics_type
  use diagnostics_dimensions, only: dim_string
  use mp, only: mp_abort,barrier
  use file_utils, only: error_unit
  use fields_parallelization, only: field_k_local
  use kt_grids, only: naky, ntheta0
  type(diagnostics_type), intent(in) :: gnostics
  integer, intent(in) :: variable_type
  character(*), intent(in) :: variable_name
  character(*), intent(in) :: dimension_list
  character(*), intent(in) :: variable_description
  character(*), intent(in) :: variable_units
  character(len=len(dimension_list)), dimension(len(dimension_list)) :: dimension_names
  character(len=len(dimension_list)) :: buffer
  integer :: xdim, idim, i, count, dimlistlength
  integer :: id, it, ik
  integer :: i1
  {data_type},{dimensionality} intent(in out) :: val
  real :: dummy

  ! Check the dimension list contains kx,ky
  ! This may need changing if we change the string used to identify kx and ky.
  xdim = index(dimension_list, trim(dim_string([gnostics%dims%kx,gnostics%dims%ky])))

  if (xdim .eq. 0) then
     write(error_unit(), *) "The function create_and_write_dstrb_field_like_variable should &
          & only be called for arrays whose dimension list contains kx,ky in that order"
     call mp_abort("")
  end if

  ! Split the dimension list into an array of dimension names
  idim = 0
  count = 1
  buffer = ''
  dimlistlength = len(trim(dimension_list))
  do i = 1,dimlistlength
    if (dimension_list(i:i) .eq. "," .or. (i .eq. dimlistlength)) then
      if (i .eq. dimlistlength) then
        buffer(count:count) = dimension_list(i:i)
        count = count + 1
      end if
      idim = idim + 1
      dimension_names(idim) = trim(buffer)
      if (trim(dimension_names(idim)) .eq. trim(dim_string(gnostics%dims%kx))) xdim = idim
      buffer = ''
      count = 1
    else
      buffer(count:count) = dimension_list(i:i)
      count = count + 1
    end if
  end do

  if (gnostics%create) then
     call create_variable(gnostics%sfile, variable_type, variable_name, trim(dimension_list), variable_description, variable_units)
     if (gnostics%distributed) then
        ! Currently not handled
     end if
  end if

  if (gnostics%wryte .or. gnostics%reed) then
     if (.not.  gnostics%distributed) then
        if (gnostics%wryte) call write_variable(gnostics%sfile, variable_name, val)
        if (gnostics%reed) call read_variable(gnostics%sfile, variable_name, val)
     else
        ! For some reason every process has to make at least
        ! one write to a variable with an infinite dimension.
        ! Here we make some dummy writes to satisfy that
        ! This will need changing if we allow multi-character dimension names
        do id = 1,idim
           if (trim(dimension_names(id)) .eq. trim(dim_string(gnostics%dims%time))) cycle
           call set_count(gnostics%sfile, variable_name, trim(dimension_names(id)), 1)
        end do
        if (gnostics%wryte) call write_variable(gnostics%sfile, variable_name, dummy)
        do id = 1,idim
           !! Reset the starts and counts
           if (trim(dimension_names(id)) .eq. trim(dim_string(gnostics%dims%time))) cycle
           call set_count(gnostics%sfile, variable_name, trim(dimension_names(id)), -1)
        end do
        call barrier

        call set_count(gnostics%sfile, variable_name, trim(dim_string(gnostics%dims%kx)), 1)
        call set_count(gnostics%sfile, variable_name, trim(dim_string(gnostics%dims%ky)), 1)
        call set_independent(gnostics%sfile, variable_name)

        do ik = 1,naky
           do it = 1,ntheta0
              if (field_k_local(it,ik)) then
                 call set_start(gnostics%sfile, variable_name, trim(dim_string(gnostics%dims%kx)), it)
                 call set_start(gnostics%sfile, variable_name, trim(dim_string(gnostics%dims%ky)), ik)
                 ! Now we treat cases where kx and ky are not the two most
                 ! slowly varying indices
                 if (xdim < {dimension} - 1) then
                    call set_count(gnostics%sfile, variable_name, trim(dimension_names(xdim+2)), 1)
                    ! This loop will normally be over species
                    do i1 = 1,size(val, xdim+2)
                       call set_start(gnostics%sfile, variable_name, trim(dimension_names(xdim+2)), i1)
                       if (xdim < {dimension} - 2) then
                          call mp_abort("Case with two dimensions to the right of kx and ky not implemented", .true.)
                       else
                          if (gnostics%wryte) call write_variable_with_offset(gnostics%sfile, variable_name, val)
                          if (gnostics%reed) call read_variable_with_offset(gnostics%sfile, variable_name, val)
                       end if
                    end do
                 else
                    if (gnostics%wryte) call write_variable_with_offset(gnostics%sfile, variable_name, val)
                    if (gnostics%reed) call read_variable_with_offset(gnostics%sfile, variable_name, val)
                 end if
              end if
           end do
        end do
        call set_collective(gnostics%sfile, variable_name)
     end if
  end if

end subroutine create_and_write_dstrb_fieldlike_variable_{id}
"""
    routine_id = get_basic_id(data_type, dimensions)

    if dimensions == 0 :
        dimensionality = ""
    else:
        dimensionality = " dimension("+(":, "*(dimensions-1))+":),"

    return routine_string.format(
        id = routine_id,
        dimensionality = dimensionality,
        data_type = data_type,
        dimension = dimensions
    ).strip()


if __name__ == "__main__":
    from itertools import product
    import argparse
    parser = argparse.ArgumentParser(
        description='Generate new diagnostics code for dealing with creating and writing.')
    parser.add_argument("output")
    args = parser.parse_args()

    create_and_write_interface_string = """
interface create_and_write_{type}
  {contents}
end interface create_and_write_{type}
"""

    data_types = ['real(kind=real32)', 'real(kind=real64)', 'integer',
                  'character', 'complex(kind=real32)', 'complex(kind=real64)']
    dimensions = [0, 1, 2, 3, 4]

    main_cases = list(product(data_types, dimensions))
    field_cases = list(product(data_types, dimensions[2:]))

    create_and_write_variable_interface = create_and_write_interface_string.format(
        contents = "\n".join([
            "  module procedure create_and_write_variable_" +
            get_basic_id(*case) for case in main_cases
            ]).strip(),
        type = 'variable'
        )

    create_and_write_variable_noread_interface = create_and_write_interface_string.format(
        contents = "\n".join([
            "  module procedure create_and_write_variable_" +
            get_basic_id(*case)+"_nr" for case in main_cases
            ]).strip(),
        type = 'variable_noread'
        )


    create_and_write_distributed_fieldlike_variable_interface = create_and_write_interface_string.format(
        contents = "\n".join([
            "  module procedure create_and_write_dstrb_fieldlike_variable_" +
            get_basic_id(*case) for case in field_cases
            ]).strip(),
        type = 'distributed_fieldlike_variable'
        )


    create_and_write_variable_implementations = "\n\n".join(
        [generate_create_and_write_variable(*case, is_no_read = False) for case in main_cases]
    )

    create_and_write_variable_noread_implementations = "\n\n".join(
        [generate_create_and_write_variable(*case, is_no_read = True) for case in main_cases]
    )


    create_and_write_distributed_fieldlike_variable_implementations = "\n\n".join(
        [generate_create_and_write_distributed_fieldlike_variable(*case) for case in field_cases]
    )

    with open(args.output, 'w') as output:
        output.write(module_string.format(
            header = header.strip(),
            create_and_write_variable_interface = create_and_write_variable_interface,
            create_and_write_variable_noread_interface = create_and_write_variable_noread_interface,
            create_and_write_distributed_fieldlike_variable_interface = create_and_write_distributed_fieldlike_variable_interface,
            create_and_write_variable_implementations = create_and_write_variable_implementations,
            create_and_write_variable_noread_implementations = create_and_write_variable_noread_implementations,
            create_and_write_distributed_fieldlike_variable_implementations = create_and_write_distributed_fieldlike_variable_implementations
            )
        )
