# This Python script automatically generates the module
# overrides. It is automatically invoked by the Makefile;
# thus, after editing it a simple make will trigger the generation of
# the file overrides.py
#
# This could make good use of jinja templating, but we do it in a more
# manual approach currently in order to avoid external dependencies

header = """
! DO NOT EDIT THIS FILE
! This file has been automatically generated using generate_overrides.py

!> A module which defines the override types. These types
!> are used within the init object (which itself is contained
!> within the gs2_program_state object) to override values
!> of the specified parameters (i.e. modify their values from
!> what is specified in the input file). The appropriate "prepare_..."
!> function from gs2_main must always be called before setting overrides.
!>
!> In the future it would be nice to replace this code with something which
!> uses the config types directly.
""".strip()

module_string = """
{header}
module overrides
  implicit none

{public_statements}
{init_values_type_declaration}

{type_declarations}

contains

{methods}

{init_values_methods}

end module overrides
""".strip()

# The initial values overrides do not follow the same pattern as the other
# override types so we do not generate these and instead hard code them here.

init_values_type_declaration = """
public :: initial_values_overrides_type

!> A type for storing overrides of the intial
!> values of the fields and distribution function.
!> This override is different to all the others,
!> because in order to minimise memory usage and disk writes,
!> this override is used internally during the simulation, and
!> its values can thus change over the course of the simulation.
!> In contrast, no other overrides are modified by running gs2.
!> Also, depending on the value of in_memory, the override
!> values will either be taken from the the arrays within
!> the object, or from the restart files. If you want
!> to externally modify the initial field and dist fn values,
!> you need to use in_memory = .true. If you just want to
!> use this override to allow you to reinitialise the equations
!> and start from the same values, you can use either memory
!> or restart files. If you want to want to change the number
!> of processors and then reinitialise and then use this override
!> you must use in_memory = .false., because currently the memory
!> is allocated on a processor by processor basis. Changing
!> grid sizes and then using this override is not supported.
!> This one is too complicated to generate
!> automatically
type initial_values_overrides_type
  !> True if the object has been initialized.
  logical, private :: init = .false.
  !> If true, override values are read from the
  !> arrays in this object. If not, they are read
  !> from the restart files. The value of in_memory
  !> should not be changed without reinitializing this
  !> object (doing so is an excellent way of generating
  !> segmentation faults).
  logical :: in_memory = .true.
  !> Whether to override initial values or not,
  !> i.e., whether or not this override is switched on.
  !> If it is switched on, initial values will be determined
  !> by the values in the arrays or the restart files,
  !> depending on the value of in_memory. If false,
  !> initial values will be determined by the gs2 input file
  !> (note that of course, this can result in initial values
  !> being taken from the input files).
  logical :: override = .false.
  ! The force_maxwell_reinit component has been removed to avoid problems in gs2_init. See comments there and in gs2_main for details.
  complex, dimension (:,:,:), allocatable :: phi
  complex, dimension (:,:,:), allocatable :: apar
  complex, dimension (:,:,:), allocatable :: bpar
  complex, dimension (:,:,:), allocatable :: g
  complex, dimension (:,:,:), allocatable :: gexp_1, gexp_2, gexp_3
  real, dimension(2) :: vnmult
 contains
  procedure :: initialise => init_initial_values_overrides
  procedure :: finish => finish_initial_values_overrides
  procedure :: is_initialised => is_initialised_initial_values_overrides
end type initial_values_overrides_type
""".strip()

init_init_values = """
!> Warning: You can't change the value of
!> overrides%in_memory after calling this function
subroutine init_initial_values_overrides(overrides_obj, ntgrid, ntheta0, naky, &
   g_llim, g_ulim, force_maxwell_reinit, in_memory, has_explicit_terms)
  use file_utils, only: error_unit
  use mp, only: proc0
  implicit none
  class(initial_values_overrides_type), intent(in out) :: overrides_obj
  integer, intent(in) :: ntgrid, ntheta0, naky, g_llim, g_ulim
  logical, intent(in) :: force_maxwell_reinit, in_memory, has_explicit_terms
  integer :: iostat
  if (overrides_obj%init) return
  overrides_obj%init = .true.
  overrides_obj%override = .false.
  overrides_obj%in_memory = in_memory
  overrides_obj%vnmult = 0.0

  if (proc0) write (error_unit(), *) "INFO: changing in_memory &
    & after calling initial_values_overrides_type will almost certainly cause &
    & segmentation faults."
  if (overrides_obj%in_memory) then
     allocate(overrides_obj%g(-ntgrid:ntgrid,2,g_llim:g_ulim), stat=iostat)
     if (has_explicit_terms) then
        if (iostat.eq.0) allocate(overrides_obj%gexp_1(-ntgrid:ntgrid,2,g_llim:g_ulim), stat=iostat)
        if (iostat.eq.0) allocate(overrides_obj%gexp_2(-ntgrid:ntgrid,2,g_llim:g_ulim), stat=iostat)
        if (iostat.eq.0) allocate(overrides_obj%gexp_3(-ntgrid:ntgrid,2,g_llim:g_ulim), stat=iostat)
     end if

     if (.not. force_maxwell_reinit) then
        if (iostat.eq.0) allocate(overrides_obj%phi(-ntgrid:ntgrid,ntheta0,naky), stat=iostat)
        if (iostat.eq.0) allocate(overrides_obj%apar(-ntgrid:ntgrid,ntheta0,naky), stat=iostat)
        if (iostat.eq.0) allocate(overrides_obj%bpar(-ntgrid:ntgrid,ntheta0,naky), stat=iostat)
     end if

     if (iostat.ne.0) then
        overrides_obj%in_memory = .false.
        if (proc0) write(error_unit(), '(A," ",I0)') "WARNING: could not allocate memory for initial_values_overrides. Only restart from file possible (manual setting of initial values not possible). Error code",iostat
        if (allocated(overrides_obj%g)) deallocate(overrides_obj%g)
        if (allocated(overrides_obj%gexp_1)) deallocate(overrides_obj%gexp_1)
        if (allocated(overrides_obj%gexp_2)) deallocate(overrides_obj%gexp_2)
        if (allocated(overrides_obj%gexp_3)) deallocate(overrides_obj%gexp_3)
        if (allocated(overrides_obj%phi)) deallocate(overrides_obj%phi)
        if (allocated(overrides_obj%apar)) deallocate(overrides_obj%apar)
        if (allocated(overrides_obj%bpar)) deallocate(overrides_obj%bpar)
     end if
  end if
end subroutine init_initial_values_overrides
""".strip()

is_init_init_values = """
elemental logical function is_initialised_initial_values_overrides(self) result(init)
  class(initial_values_overrides_type), intent(in) :: self
  init = self%init
end function is_initialised_initial_values_overrides
""".strip()

finish_init_values = """
subroutine finish_initial_values_overrides(overrides_obj)
  use file_utils, only: error_unit
  use mp, only: proc0
  implicit none
  class(initial_values_overrides_type), intent(in out) :: overrides_obj
  if (.not. overrides_obj%init) then
    if (proc0) write (error_unit(), *) "WARNING: Called finish_initial_values_overrides on an uninitialized object"
    return
  end if
  overrides_obj%init = .false.
  overrides_obj%override = .false.
  overrides_obj%vnmult = 0.0
  if (allocated(overrides_obj%g)) deallocate(overrides_obj%g)
  if (allocated(overrides_obj%gexp_1)) deallocate(overrides_obj%gexp_1)
  if (allocated(overrides_obj%gexp_2)) deallocate(overrides_obj%gexp_2)
  if (allocated(overrides_obj%gexp_3)) deallocate(overrides_obj%gexp_3)
  if (allocated(overrides_obj%phi)) deallocate(overrides_obj%phi)
  if (allocated(overrides_obj%apar)) deallocate(overrides_obj%apar)
  if (allocated(overrides_obj%bpar)) deallocate(overrides_obj%bpar)
end subroutine finish_initial_values_overrides
""".strip()

# Here we declare some helpers for dealing with the generic override types
type_declaration_boilerplate = """
{doc_string}
type {name}_overrides_type
  logical, private :: init = .false.
{parameters}
 contains
  procedure :: initialise => init_{name}_overrides
  procedure :: finish => finish_{name}_overrides
  procedure :: is_initialised => is_initialised_{name}_overrides
end type {name}_overrides_type
""".rstrip()

init_method_boilerplate = """
subroutine init_{name}_overrides({obj}{extra_args})
  class({name}_overrides_type), intent(in out) :: {obj}
  {extra_decls}
  if ({obj}%init) return
  {obj}%init = .true.
{parameter_inits}
end subroutine init_{name}_overrides
""".rstrip()

is_initialised_method_boilerplate = """
elemental logical function is_initialised_{name}_overrides({obj}) result(init)
  class({name}_overrides_type), intent(in) :: {obj}
  init = {obj}%init
end function is_initialised_{name}_overrides
""".rstrip()

finish_method_boilerplate = """
subroutine finish_{name}_overrides({obj})
  use file_utils, only: error_unit
  class({name}_overrides_type), intent(in out) :: {obj}
  if (.not. {obj}%init) then
    write (error_unit(), *) "ERROR: Called finish_{name}_overrides on an uninitialized object."
    return
  end if
{parameter_finish}
end subroutine finish_{name}_overrides
""".rstrip()

class override_parameter():
    """ Small helper to represent a type to override """
    def __init__(self, data_type, name, size = None, default = None):
        self.data_type = data_type
        self.name = name
        self.size = size
        self.default = default

    def override_flag(self):
        if self.size is not None:
            # Note we currently assume any parameters with a size are 1d.
            # It should be easy to generalise this to ND by taking a list/tuple
            # and determining the size from there.
            size_string = ", dimension (:), allocatable"
            default_string = ""
        else:
            size_string = ""
            default_string = " = .false."

        return "logical{size} :: override_{name}{default}".format(name = self.name, size = size_string, default = default_string)

    def var_declaration(self):
        if self.default is not None:
            default_string = " = {default}".format(default=self.default)
        else:
            default_string = ""

        if self.size is not None:
            # Note we currently assume any parameters with a size are 1d.
            # It should be easy to generalise this to ND by taking a list/tuple
            # and determining the size from there.
            size_string = ", dimension (:), allocatable"
        else:
            size_string = ""

        return "{data_type}{size} :: {name}{default}".format(
            data_type = self.data_type,
            size = size_string,
            name = self.name,
            default =default_string
        )

    def declaration(self):
        return "\n".join([self.override_flag(), self.var_declaration()]).rstrip()

    def init_statement(self, object_name = "self"):
        if self.size is not None:
            init_string = "\n".join([
                "allocate({obj}%override_{name}({size}), {obj}%{name}({size}))".format(
                    obj = object_name, name = self.name, size = self.size),
                "{obj}%override_{name} = .false.".format(obj = object_name, name = self.name)
                ])
        else:
            init_string = ""

        return init_string

    def finish_statement(self, object_name = "self"):
        if self.size is not None:
            finish_string = "if ( allocated({obj}%override_{name})) deallocate({obj}%override_{name}, {obj}%{name})".format(obj = object_name, name = self.name)
        else:
            finish_string = "{obj}%override_{name} = .false.".format(obj = object_name, name = self.name)

        return finish_string

class override_set():
    """ Helper type to represent a set of parameters to override """
    def __init__(self, name, parameters, doc_string = None):
        self.name = name
        self.parameters = parameters

        if doc_string is None:
            self.doc_string = "!> An object for overriding {name} parameters".format(name = self.name)
        else:
            self.doc_string = doc_string

        # Get all size names associated with parameters which aren't none
        self.size_names = set([x.size for x in self.parameters if x.size is not None])

    def declaration(self):
        return type_declaration_boilerplate.format(
            doc_string = self.doc_string.rstrip(),
            name = self.name,
            parameters = "\n".join([x.declaration() for x in self.parameters]).rstrip()
            ).rstrip()

    def public(self):
        return "public :: {name}_overrides_type".format(name = self.name).rstrip()

    def init_method(self):
        if len(self.size_names) > 0:
            extra_args = ", " + ", ".join(self.size_names)
            extra_decls = "\n".join(["integer, intent(in) :: {size}".format(
                size = x) for x in self.size_names])
        else:
            extra_args = ""
            extra_decls = ""

        return init_method_boilerplate.format(
            name = self.name,
            obj = "self",
            extra_args = extra_args,
            extra_decls = extra_decls,
            parameter_inits = "\n".join([x.init_statement() for x in self.parameters]).rstrip()
            ).rstrip()

    def finish_method(self):
        return finish_method_boilerplate.format(
            name = self.name,
            obj = "self",
            parameter_finish = "\n".join([x.finish_statement() for x in self.parameters]).rstrip()
            ).rstrip()

    def is_initialised_method(self):
        return is_initialised_method_boilerplate.format(name = self.name, obj = "self").rstrip()

    def method(self):
        return "\n".join([self.init_method(), self.is_initialised_method(), self.finish_method()]).rstrip()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description='Generate overrides code for dealing with overriding various settings.')
    parser.add_argument("output")
    args = parser.parse_args()

    profiles = override_set(
        name = "profiles",
        parameters = [
            override_parameter("real", "dens", "nspec"),
            override_parameter("real", "temp", "nspec"),
            override_parameter("real", "tprim", "nspec"),
            override_parameter("real", "fprim", "nspec"),
            override_parameter("real", "vnewk", "nspec"),
            override_parameter("real", "g_exb"),
            override_parameter("real", "mach")
            ],
        doc_string = """
!> An object for overriding all or selected
!> profile parameters, for example species
!> temps, densities or gradients or the flow gradient or mach
!> number. Note that all species parameters are arrays of
!> size nspec and you must set the override switches
!> individually for each species.
"""
        )

    miller = override_set(
        name = "miller_geometry",
        parameters =  [
            override_parameter("real", "rhoc"),
            override_parameter("real", "qinp"),
            override_parameter("real", "shat"),
            override_parameter("real", "rgeo_lcfs"),
            override_parameter("real", "rgeo_local"),
            override_parameter("integer", "geoType"),
            override_parameter("real", "aSurf"),
            override_parameter("real", "shift"),
            override_parameter("real", "shiftVert"),
            override_parameter("integer", "mMode"),
            override_parameter("integer", "nMode"),
            override_parameter("real", "deltam"),
            override_parameter("real", "deltan"),
            override_parameter("real", "deltampri"),
            override_parameter("real", "deltanpri"),
            override_parameter("real", "thetam"),
            override_parameter("real", "thetan"),
            override_parameter("real", "betaprim"),
            override_parameter("real", "akappa"),
            override_parameter("real", "akappri"),
            override_parameter("real", "tri"),
            override_parameter("real", "tripri")
            ],
        doc_string = """
!> An object for overriding all or selected
!> Miller geometry parameters.
"""
        )

    optimisations = override_set(
        name = "optimisations",
        parameters = [
            override_parameter("integer", "nproc"),
            override_parameter("logical", "opt_redist_nbk"),
            override_parameter("logical", "opt_redist_persist"),
            override_parameter("logical", "opt_redist_persist_overlap"),
            override_parameter("logical", "intmom_sub"),
            override_parameter("logical", "intspec_sub"),
            override_parameter("logical", "local_field_solve"),
            override_parameter("character(len=5)", "layout"),
            override_parameter("character(len=8)", "field_option"),
            override_parameter("logical", "gf_lo_integrate"),
            override_parameter("logical", "gf_local_fields"),
            override_parameter("logical", "simple_gf_decomposition"),
            override_parameter("logical", "field_subgath"),
            override_parameter("logical", "do_smart_update"),
            override_parameter("logical", "field_local_allreduce"),
            override_parameter("logical", "field_local_allreduce_sub"),
            override_parameter("logical", "opt_source"),
            override_parameter("integer", "minnrow"),
            override_parameter("integer", "old_comm") #This is a special value that we need
            ],
        doc_string = """
!> A type for containing overrides to the processor layout
!> and optimisation flags for gs2.
"""
        )

    timestep = override_set(
        name = "timestep",
        parameters = [
            override_parameter("logical", "immediate_reset")
            ],
        doc_string = """
!> A type for containing overrides to the timestep and the cfl parameters
"""
        )

    kt_grids = override_set(
        name = "kt_grids",
        parameters = [
            override_parameter("integer", "ny"),
            override_parameter("integer", "naky"),
            override_parameter("integer", "nx"),
            override_parameter("integer", "ntheta0"),
            override_parameter("real", "y0"),
            override_parameter("real", "x0"),
            override_parameter("integer", "jtwist"),
            override_parameter("logical", "gryfx")
            ],
        doc_string = """
!> A type for containing overrides to the perpendicular grids (x and y).
"""
    )

    sets = [profiles, miller, optimisations, timestep, kt_grids]

    with open(args.output, 'w') as output:
        output.write(module_string.format(
            header = header.rstrip(),
            public_statements = "\n".join([x.public().strip() for x in sets]),
            type_declarations = "\n\n".join([x.declaration().strip() for x in sets]),
            init_values_type_declaration = init_values_type_declaration.strip(),
            methods = "\n\n".join([x.method().strip() for x in sets]),
            init_values_methods = "\n".join([init_init_values.strip(), is_init_init_values.strip(), finish_init_values.strip()]).rstrip()
            )
        )
