# This Python script automatically generates the module
# read_chease. It is NOT automatically invoked by the Makefile;
# thus, after editing it need to run the following to generate
# the updated fortran file:
#      python generate_read_chease read_chease.f90
#
# This could make good use of jinja templating, but we do it in a more
# manual approach currently in order to avoid external dependencies


header = """
! DO NOT EDIT THIS FILE
! This file has been automatically generated using generate_read_chease.py

!> CHEASE Output Reader
!>
!> A module to read in datafiles from CHEASE.
!>
!> Based on code written by Edmund Highcock
!> edmundhighcock@sourceforge.net
!>
!> This is free software released under the MIT Licence
!>
!> This module is generated automatically
!> using
!>    $ python generate_read_chease.py read_chease.f90
"""

module_string = """
{header}
module read_chease
  implicit none

  private

  public :: read_infile
  public :: finish

  public :: npsi_chease, nchi_chease

{public_statements}

  integer :: npsi_chease, nchi_chease

  ! We could either use [[file_utils:get_unused_unit]] or F2008's
  ! newunit option to open to avoid hard-coding this. We don't do
  ! this now to avoid any dependencies on the rest of gs2 or F2008.
  integer, parameter :: infile = 1212

{declarations}

  logical, parameter :: debug = .true.

contains

{read_infile}
{finish}

end module read_chease
"""

def generate_read_infile(zero_d, one_d, two_d):
    read_infile_string = """
!> Open and read chease file named "filename". Populates
!> module level variables with the output.
subroutine read_infile(filename)
  implicit none
  character(len=*), intent(in) :: filename

  open(unit = infile, file= filename)

  read(infile, *) ; read(infile, *) npsi_chease
  if (debug) write(*, *) npsi_chease, "<---npsi_chease"

  read(infile, *) ; read(infile, *) nchi_chease
  if (debug) write(*, *) nchi_chease, "<---nchi_chease"

{zero_d_read}

{one_d_read}

{two_d_read}

  close(unit = infile)
end subroutine read_infile
"""

    read_statement = "read(infile, *)"
    read_var = "  {read} ; {read} {var}"
    allocate_var = "  allocate( {var}({shape}))"
    one_d_shape = "npsi_chease"
    two_d_shape = "npsi_chease, nchi_chease"

    zero_d_read = read_var
    one_d_read = "\n".join([allocate_var, read_var])
    two_d_read = "\n".join([allocate_var, read_var])

    zero_d_reads = "\n\n".join([zero_d_read.format(
        read = read_statement,
        var = var).rstrip() for var in zero_d])

    # Need some special handling as the first two are
    # psi_chease and chi_chease with size npsi_chease and
    # nchi_chease respectively
    one_d_reads = one_d_read.format(read = read_statement, shape = one_d_shape, var = one_d[0])
    one_d_reads += "\n\n" + one_d_read.format(read = read_statement, shape = "nchi_chease", var = one_d[1]) + "\n"
    one_d_reads += "\n\n".join([one_d_read.format(
        read = read_statement,
        var = var,
        shape = one_d_shape).rstrip() for var in one_d[2:]])

    two_d_reads = "\n\n".join([two_d_read.format(
        read = read_statement,
        var = var,
        shape = two_d_shape).rstrip() for var in two_d])

    return read_infile_string.format(
        zero_d_read = zero_d_reads,
        one_d_read = one_d_reads,
        two_d_read = two_d_reads).rstrip()

def generate_finish(zero_d, one_d, two_d):
    finish_string = """
!> Deallocate any module level arrays which have been allocated
subroutine finish
  implicit none
{one_d_deallocates}
{two_d_deallocates}
end subroutine finish
"""
    deallocate = "  if ( allocated({var})) deallocate({var})"

    one_d_deallocates = "\n".join([deallocate.format(var = var) for var in one_d])
    two_d_deallocates = "\n".join([deallocate.format(var = var) for var in two_d])

    return finish_string.format(
        one_d_deallocates = one_d_deallocates,
        two_d_deallocates = two_d_deallocates).rstrip()

def generate_public(zero_d, one_d, two_d):
    public_string = "  public :: {var}"

    return "\n".join([ public_string.format(var = var) for var in
        [*zero_d, *one_d, *two_d]
        ]).rstrip()

def generate_declarations(zero_d, one_d, two_d):
    declaration_string = "  real{properties} :: {var}"
    zero_d_declarataions = "\n".join([declaration_string.format(
        properties = "",
        var = var) for var in zero_d])

    one_d_declarataions = "\n".join([declaration_string.format(
        properties = ", dimension(:), allocatable",
        var = var) for var in one_d])

    two_d_declarataions = "\n".join([declaration_string.format(
        properties = ", dimension(:, :), allocatable",
        var = var) for var in two_d])

    return "\n".join([zero_d_declarataions, one_d_declarataions, two_d_declarataions]).rstrip()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description='Generate code for reading chease equilibrium file.')
    parser.add_argument("output")
    args = parser.parse_args()

    zero_d = [ "R0EXP", "B0EXP"]
    one_d =  [ "PSI", "CHI", "Rgeom", "ageom", "q", "dqdpsi",
               "d2qdpsi2", "p", "dpdpsi", "f", "fdfdpsi", "V",
               "rho_t", "shear", "dsheardpsi", "kappa", "delta_lower",
               "delta_upper", "dVdpsi", "dpsidrhotor", "GDPSI_av",
               "radius_av", "R_av", "TE", "DTEDPSI", "NE", "DNEDPSI",
               "TI", "DTIDPSI", "NI", "DNIDPSI", "ZEFF", "SIGNEO", "JBSBAV"]
    two_d =  [ "g11", "g12", "g22", "g33", "B", "dBdpsi", "dBdchi", "dPsidR",
               "dPsidZ", "dChidR", "dChidZ", "Jacobian", "R", "Z"]

    # Convert to lower case and add "_chease"
    zero_d = [x.lower()+"_chease" for x in zero_d]
    one_d = [x.lower()+"_chease" for x in one_d]
    two_d = [x.lower()+"_chease" for x in two_d]

    the_vars = [zero_d, one_d, two_d]

    with open(args.output, 'w') as output:
        output.write(module_string.format(
            header = header.strip(),
            public_statements = generate_public(*the_vars),
            declarations = generate_declarations(*the_vars),
            read_infile = generate_read_infile(*the_vars),
            finish = generate_finish(*the_vars)
        ))
