! This file is used alongside code generated by generate_gs2_init.py

!> This module is analogous to the init() function
!> in Linux-based operating systems: it initialises
!> gs2 to a certain init_level. At a given init level,
!> certain modules are initialised and certain are not.
!>
!> The gs2_init module is used by gs2_main to initialise modules. A typical
!> additional use case for this module is when it is desired
!> to override a given parameter (as in the override_* functions
!> in gs2_main). GS2 must be taken down to the appropriate
!> init_level, where all modules which contain any of those
!> parameters are uninitialized. The override is then set
!> and gs2 is brought back up to the highest init_level.
!>
!> As in Linux, this module cannot be used until a certain
!> basic initialization has happened (think loading the kernel).
!> This basic initialization occurs in gs2_initialize in gs2_main,
!> and set the init_level to gs2_initialized.
!>
!> This is free software released under the MIT licence.
!> Originally written by:
!>            Edmund Highcock (edmundhighcock@users.sourceforge.net)
module gs2_init
  use abstract_config, only: abstract_config_type, CONFIG_MAX_NAME_LEN
  use overrides, only: miller_geometry_overrides_type, initial_values_overrides_type, optimisations_overrides_type
  use overrides, only: profiles_overrides_type, kt_grids_overrides_type, timestep_overrides_type

  implicit none

  private

  public :: init_type

  !> A list of possible intialization levels.
  public :: init_level_list

  !> Bring gs2 to the target initialization level.
  public :: init

  !> Reads the gs2_init namelist
  public :: init_gs2_init

  !> Finalize the module
  public :: finish_gs2_init

  public :: in_memory
  public :: init_config_type
  public :: set_gs2_init_config
  public :: get_gs2_init_config

  !> A type for labelling the different init
  !> levels available in gs2.
  type init_level_list_type
     !> The init_level reaches basic
     !> when initialize_gs2 has been called.
     integer :: basic = 1
#include "gs2_init_level_list.inc"
  end type init_level_list_type

  type(init_level_list_type), parameter :: init_level_list = init_level_list_type()

  !> A type for storing the current initialization
  !> status, as well as all the overrides.
  type init_type
     !> The current init level
     integer :: level = 0
     !> Whether or not diagnostics have been initialized
     !> Not obvious why this belongs here really as we don't
     !> deal with diagnostics in this module.
     logical :: diagnostics_initialized = .false.
     !> An object for overriding all or selected
     !> Miller geometry parameters. You must call
     !> gs2_main::prepare_miller_geometry_overrides
     !> before setting these overrides. See
     !> documentation for the overrides::miller_geometry_overrides_type
     !> for more information.
     type(miller_geometry_overrides_type) :: mgeo_ov
     !> An object for overriding all or selected
     !> profile parameters such as species temperature, density, and gradients
     !> as well as the flow gradient and mach number. You must call
     !> gs2_main::prepare_profiles_overrides
     !> before setting these overrides. See
     !> documentation for the overrides::profiles_overrides_type
     !> for more information.
     type(kt_grids_overrides_type) :: kt_ov
     !> An object for overriding all or selected
     !> kt_grids parameters such as ny, nx, y0, jtwist etc.
     !> You must call
     !> gs2_main::prepare_kt_grids_overrides
     !> before setting these overrides. See
     !> documentation for the overrides::kt_grids_overrides_type
     !> for more information.
     type(profiles_overrides_type) :: prof_ov
     !> An object for overriding parameters connected
     !> to the timestep and cfl condition
     type(timestep_overrides_type) :: tstep_ov
     !> An object for overriding the initial values of
     !> the fields and distribution function. You must call
     !> gs2_main::prepare_initial_values_overrides
     !> before setting these overrides. This override
     !> is very complicated. See
     !> documentation for the overrides::initial_values_overrides_type
     !> for more information.
     type(initial_values_overrides_type) :: initval_ov
     !> An object for overriding non physics parameters which
     !> may alter run time and efficiency. You must call
     !> gs2_main::prepare_optimisations_overrides
     !> before setting these overrides.
     type(optimisations_overrides_type) :: opt_ov
  end type init_type

  logical :: in_memory = .false.
  logical :: initialized = .false.

  !> Used to represent the input configuration of init
  type, extends(abstract_config_type) :: init_config_type
     ! namelist : init_knobs
     ! indexed : false
     !> If `true` then when changing the timestep attempt to keep a
     !> copy of the current fields and distribution function in memory
     !> rather than dumping to disk. This should fall back to a disk
     !> based method if there is insufficient memory. Currently both
     !> this value and the value of `in_memory` in [[reinit_knobs]]
     !> must be set to `true` in order to enable this feature.
     !>
     !> @todo Consider if this can be removed and just the
     !> [[reinit_knobs]] version controls the feature.
     logical :: in_memory = .false.
   contains
     procedure, public :: read => read_init_config
     procedure, public :: write => write_init_config
     procedure, public :: reset => reset_init_config
     procedure, public :: broadcast => broadcast_init_config
     procedure, public, nopass :: get_default_name => get_default_name_init_config
     procedure, public, nopass :: get_default_requires_index => get_default_requires_index_init_config
  end type init_config_type

  type(init_config_type) :: init_config

contains
  !> Initialize gs2 to the level of target_level.
  !> The init_type current contains info
  !> about the current initialization level. At the end
  !> of the subroutine, current%level is set to target_level
  subroutine init(current, target_level)
    use fields, only: init_fields
    use mp, only: mp_abort
    implicit none
    type(init_type), intent(inout) :: current
    integer, intent(in) :: target_level

    if (current%level .lt. init_level_list%basic) then
       call mp_abort("gs2_init::init cannot be called before &
            & initialize_gs2 in gs2 main", .true.)
    end if

    if (current%level .eq. target_level) then
       return
    else
       if (up()) then
#include "gs2_init_up.inc"
       else if (down()) then
#include "gs2_init_down.inc"
       end if
    end if
  contains
    pure function up()
      logical :: up
      up = (target_level>current%level)
    end function up
    pure function down()
      logical :: down
      down = (target_level<current%level)
    end function down
#include "gs2_init_subroutines.inc"
  end subroutine init

  !> Initialise this module
  subroutine init_gs2_init(current, init_config_in)
    implicit none
    type(init_type), intent(inout) :: current
    type(init_config_type), intent(in), optional :: init_config_in

    logical :: exist
    if (initialized) return
    initialized = .true.

    if (present(init_config_in)) init_config = init_config_in

    call init_config%init(name = 'init_knobs', requires_index = .false.)

    ! Copy out internal values into module level parameters
    in_memory = init_config%in_memory

    exist = init_config%exist
  end subroutine init_gs2_init

  !> Finish this module
  subroutine finish_gs2_init(current)
    type(init_type), intent(inout) :: current
    call init_config%reset()
    initialized = .false.
  end subroutine finish_gs2_init

  subroutine set_initial_field_and_dist_fn_values(current)
    use dist_fn_arrays, only: g, gnew, gexp_1, gexp_2, gexp_3
    use fields_arrays, only: phinew, aparnew, bparnew, phi, apar, bpar
    use fields, only: force_maxwell_reinit
    use fields, only: set_init_fields
    use file_utils, only: error_unit
    use init_g, only: ginit, init_vnmult
    use init_g, only: ginitopt_restart_many, initial_condition_is_nonadiabatic_dfn
    use run_parameters, only: fphi, fapar, fbpar
    use collisions, only: vnmult
    implicit none
    type (init_type), intent(in) :: current
    logical :: restarted

    if (.not. current%initval_ov%override) then
       ! This is the usual initial setup
       call ginit (restarted)
       ! If initial_condition_is_nonadiabatic_dfn then we have already
       ! calculated the correct fields, unless restarting, so return straight away.
       ! If restarting then we deal with this further below.
       if ((.not.restarted) .and. initial_condition_is_nonadiabatic_dfn) return
       ! If we're restarting from file then we don't want to recalculate
       ! the fields unless force_maxwell_reinit is true. If we are not
       ! restarting then we better call set_init_fields to set the initial
       ! fields here.
       if ((.not.restarted) .or. force_maxwell_reinit) call set_init_fields
       return
    else
       if (current%initval_ov%in_memory) then
          g = current%initval_ov%g
          gnew = g
          vnmult = current%initval_ov%vnmult
       else
          call ginit(restarted, ginitopt_restart_many)
          call init_vnmult(vnmult)
       end if
    end if

    if (current%initval_ov%in_memory) then
       if (allocated(current%initval_ov%gexp_1)) gexp_1 = current%initval_ov%gexp_1
       if (allocated(current%initval_ov%gexp_1)) gexp_2 = current%initval_ov%gexp_2
       if (allocated(current%initval_ov%gexp_1)) gexp_3 = current%initval_ov%gexp_3
    end if

    ! Do not use the value from state%init%initval_ov%force_maxwell_reinit = current%initval_ov%force_maxwell_reinit as follows:
    !     if (current%initval_ov%force_maxwell_reinit)then
    ! as was done previously because, when restarting a run, this part of the overrides system has not been initialised with the
    ! value from the input file yet. As a result, the above condition is always true, even if the restart file has
    ! force_maxwell_reinit = .false.! (For a reinitialisation due to a timestep change, this part of the override system would be
    ! set up correctly and the above condition would respect the value in the input file.) Unfortunately, because of the way the
    ! state object and the overrides system is set up, it is not possible to have this value correctly initialised from the input
    ! file by this point in a restarted run. Therefore, use the input file value directly to ensure we respect the value in the
    ! input file at the start of restarted runs.
    if (force_maxwell_reinit)then
       call set_init_fields
    else

       if(current%initval_ov%override .and. current%initval_ov%in_memory) then
          if(fphi.gt.0) then
             phinew=current%initval_ov%phi
          else
             phinew = 0.
          endif
          if(fapar.gt.0) then
             aparnew=current%initval_ov%apar
          else
             aparnew = 0.
          endif
          if(fbpar.gt.0) then
             bparnew=current%initval_ov%bpar
          else
             bparnew = 0.
          endif
          phi = phinew
          apar = aparnew
          bpar = bparnew
       else
          ! No need to do anything: fields read from file in
          ! [[gs2_save:gs2_restore_many]]
       end if
    end if
  end subroutine set_initial_field_and_dist_fn_values

  !> Set the module level config type
  !> Will abort if the module has already been initialised to avoid
  !> inconsistencies.
  subroutine set_gs2_init_config(gs2_init_config_in)
    use mp, only: mp_abort
    type(init_config_type), intent(in), optional :: gs2_init_config_in
    if (initialized) then
       call mp_abort("Trying to set gs2_init_config when already initialized.", to_screen = .true.)
    end if
    if (present(gs2_init_config_in)) then
       init_config = gs2_init_config_in
    end if
  end subroutine set_gs2_init_config

  !> Get the module level config instance
  function get_gs2_init_config()
    type(init_config_type) :: get_gs2_init_config
    get_gs2_init_config = init_config
  end function get_gs2_init_config

  !---------------------------------------
  ! Following is for the config_type
  !---------------------------------------

  !> Reads in the init_knobs namelist and populates the member variables
  subroutine read_init_config(self)
    use file_utils, only: input_unit_exist, get_indexed_namelist_unit
    use mp, only: proc0
    implicit none
    class(init_config_type), intent(in out) :: self
    logical :: exist
    integer :: in_file

    ! Note: When this routine is in the module where these variables live
    ! we shadow the module level variables here. This is intentional to provide
    ! isolation and ensure we can move this routine to another module easily.
    logical :: in_memory

    namelist /init_knobs/ in_memory

    ! Only proc0 reads from file
    if (.not. proc0) return

    ! First set local variables from current values
    in_memory = self%in_memory

    ! Now read in the main namelist
    in_file = input_unit_exist(self%get_name(), exist)
    if (exist) read(in_file, nml = init_knobs)

    ! Now copy from local variables into type members
    self%in_memory = in_memory

    self%exist = exist
  end subroutine read_init_config

  !> Writes out a namelist representing the current state of the config object
  subroutine write_init_config(self, unit)
    implicit none
    class(init_config_type), intent(in) :: self
    integer, intent(in) , optional:: unit
    integer :: unit_internal

    unit_internal = 6 ! @todo -- get stdout from file_utils
    if (present(unit)) then
       unit_internal = unit
    endif

    call self%write_namelist_header(unit_internal)
    call self%write_key_val("in_memory", self%in_memory, unit_internal)
    call self%write_namelist_footer(unit_internal)
  end subroutine write_init_config

  !> Resets the config object to the initial empty state
  subroutine reset_init_config(self)
    class(init_config_type), intent(in out) :: self
    type(init_config_type) :: empty
    select type (self)
    type is (init_config_type)
       self = empty
    end select
  end subroutine reset_init_config

  !> Broadcasts all config parameters so object is populated identically on
  !! all processors
  subroutine broadcast_init_config(self)
    use mp, only: broadcast
    implicit none
    class(init_config_type), intent(in out) :: self
    call broadcast(self%in_memory)

    call broadcast(self%exist)
  end subroutine broadcast_init_config

  !> Gets the default name for this namelist
  function get_default_name_init_config()
    implicit none
    character(len = CONFIG_MAX_NAME_LEN) :: get_default_name_init_config
    get_default_name_init_config = "init_knobs"
  end function get_default_name_init_config

  !> Gets the default requires index for this namelist
  function get_default_requires_index_init_config()
    implicit none
    logical :: get_default_requires_index_init_config
    get_default_requires_index_init_config = .false.
  end function get_default_requires_index_init_config
end module gs2_init
