
!> A program that runs unit tests on the theta_grid module.
!!
!! Current tests:
!!   1. A given geometry is provided in Miller, EFIT and CHEASE form. We test 
!!       to see that theta_grid gets the same coefficients in all three cases.
!!
!!
!! This is free software released under the MIT license
!!   Written by: Edmund Highcock (edmundhighcock@users.sourceforge.net)
program test_theta_grid

  use unit_tests
  use theta_grid, only: init_theta_grid, finish_theta_grid
  use theta_grid, only: initialized, theta
  use theta_grid, only: bmag, grho, gbdrift, cvdrift, gds2, gds21
  use egrid
  use mp, only: init_mp, finish_mp, proc0
  use file_utils, only: init_file_utils, finish_file_utils

  implicit none
  real :: eps
  logical :: dummy
  real, dimension(33, 3, 10) :: rslt
  real, dimension(33, 10) :: mill_rslt
  real, dimension(33) :: theta_rslt


  ! General config
  eps = 1.0e-7

  if (precision(eps).lt. 11) eps = eps * 1000.0
  !if (precision(eps).lt. 11) write (*,*) 'low precision'

  ! Set up depenencies
  call init_mp
  if (proc0) call init_file_utils(dummy, .true., .true., .false., 'local')
  !call init_species
  !call init_general_f0



  call announce_module_test('theta_grid')

  !call announce_test('init_theta_grid')

  !call process_test(theta_grid_unit_test_init_theta_grid(sizes, energy_results, eps), &
    !'init_theta_grid')
  !call init_theta_grid(dummy, dummy)

  !call init_dist_fn
  call init_theta_grid
  !write (*,*) 'size', size(bmag)

  mill_rslt = miller_result()

  rslt = 0.
  call get_results(1)
!
  !call finish_theta_grid
  if (proc0) call finish_file_utils
  !write(*,*) 'MILLER', rslt(:,1,:)
  
  !write (*,*) 'Finished local'

  !initialized = .false.

  if (proc0) call init_file_utils(dummy, .true., .true., .false., 'chease')
  call get_results(2)
  if (proc0) call finish_file_utils

  !initialized = .false.

  !theta_rslt = - theta_rslt

  if (proc0) call init_file_utils(dummy, .true., .true., .false., 'efit')
  call get_results(3)
  if (proc0) call finish_file_utils


  call announce_test(' that Miller has not changed')
  !rslt(:,1,1) = rslt(:, 1, 1) * 0.9901928539149342 !* 0.9934 ! Adjust for different normalisation in
  ! Miller

  call process_test( &
    test_results(rslt(:,1,:), miller_result(), 1.0e-2, 1.0e-2), &
    ' that Miller has not changed')

  call announce_test(' that Miller and CHEASE give the same result')
  !rslt(:,1,1) = rslt(:, 1, 1) * 0.9901928539149342 !* 0.9934 ! Adjust for different normalisation in
  ! Miller

  call process_test( &
    test_results(rslt(:,1,:), rslt(:,2,:), 4.0e-1, 1.1e-2), &
    ' that Miller and CHEASE give the same result')

  call announce_test(' that CHEASE and EFIT give the same result')
  !rslt(:,1,1) = rslt(:, 1, 1) * 0.9901928539149342 !* 0.9934 ! Adjust for different normalisation in
  ! Miller

  call process_test( &
    test_results(rslt(:,2,:), rslt(:,3,:), 27.0e-2, 1.0e-1), &
    ' that CHEASE and EFIT give the same result')

  call close_module_test('theta_grid')
  call finish_mp

contains
  function test_results(rslt1, rslt2, epsin, bmagerror)
    real, dimension(:,:) :: rslt1, rslt2
    logical :: test_results
    real :: epsin, bmagerror
    test_results = .true.
    call announce_check('bmag')
    call process_check(test_results, &
      agrees_with(rslt1(:,1), rslt2(:,1), bmagerror), 'bmag')
    call announce_check('grho')
    call process_check(test_results, &
      agrees_with(rslt1(:,2), rslt2(:,2), bmagerror), 'grho')
    call announce_check('gbdrift')
    call process_check(test_results, &
      agrees_with(rslt1(:,3), rslt2(:,3), epsin), 'gbdrift')
    call announce_check('cvdrift')
    call process_check(test_results, &
      agrees_with(rslt1(:,4), rslt2(:,4), epsin), 'cvdrift')
    !call announce_check('theta')
    !call process_check(test_results, &
      !agrees_with(rslt1(:,7), rslt2(:,7), epsin), 'theta')

  end function test_results
  subroutine get_results(i)
    integer, intent(in) :: i
    call init_theta_grid
    if (i .eq. 1) theta_rslt = mill_rslt(:,7)
    !write (*,*) 'agwith', agrees_with(theta, theta_rslt, 0.1) 
    !theta_rslt = 1.0
    rslt(:,i,1) = spl_fit(bmag)
    rslt(:,i,2) = spl_fit(grho)
    rslt(:,i,3) = spl_fit(gbdrift)
    !write (*,*) 'gbdrift', gbdrift, 'rslt', rslt(:,i,3)
    rslt(:,i,4) = spl_fit(cvdrift)
    rslt(:,i,5) = spl_fit(gds2)
    rslt(:,i,6) = spl_fit(gds21)
    !rslt(:,i,7) = spl_fit(theta)
    !rslt(:,i,2) = bmag
    call finish_theta_grid
  end subroutine get_results

  function spl_fit(var)
    use splines
    type(spline) :: spl
    real, dimension(:), intent(in) :: var
    real, dimension(33) :: spl_fit
    integer :: i
    !write (*,*) 'Size of theta = ', size(theta)
    !write (*,*) 'Size of var = ', size(var)
    !write (*,*) 'Size of theta_rslt = ', size(theta_rslt)
    call new_spline(size(var), theta, var, spl)
    do i = 1,33
      spl_fit(i) = splint(theta_rslt(i), spl)
    end do
    call delete_spline(spl)
  end function spl_fit

  function miller_result()
    real, dimension (33,10) :: miller_result

miller_result = &
reshape([&
!Bmag
0.990894104566828, &
0.990230629474429, &
0.988289540740952, &
0.985100674218816, &
0.980756384123103, &
0.975351567198393, &
0.968967947082702, &
0.961876871776874, &
0.954114958294577, &
0.93734366136929, &
0.919933653576351, &
0.903392317922455, &
0.888524549211204, &
0.87620532470962, &
0.867104363900617, &
0.861492615953639, &
0.859597143030485, &
0.861492615953639, &
0.867104363900617, &
0.87620532470962, &
0.888524549211204, &
0.903392317922455, &
0.919933653576351, &
0.93734366136929, &
0.954114958294577, &
0.961876871776874, &
0.968967947082702, &
0.975351567198393, &
0.980756384123103, &
0.985100674218816, &
0.988289540740952, &
0.990230629474429, &
0.990894104566828, &
!GRHO
0.781943648575045, &
0.780220596004828, &
0.776014770429850, &
0.768748270605451, &
0.759233386885267, &
0.748042391277348, &
0.735946250614388, &
0.724484226863444, &
0.715182272410294, &
0.712665719710369, &
0.750320156259619, &
0.841972908689019, &
0.979948892394762, &
1.134030022532600, &
1.267561369300460, &
1.356063294319460, &
1.386702010361190, &
1.356063294319460, &
1.267561369300460, &
1.134030022532600, &
0.979948892394760, &
0.841972908689019, &
0.750320156259619, &
0.712665719710369, &
0.715182272410294, &
0.724484226863444, &
0.735946250614388, &
0.748042391277347, &
0.759233386885267, &
0.768748270605451, &
0.776014770429850, &
0.780220596004828, &
0.781943648575045, &
!GBDRIFT
-0.747708252680297, &
-0.706371081300484, &
-0.664795100033688, &
-0.622501288563042, &
-0.579323980561703, &
-0.534128406327253, &
-0.486008290350662, &
-0.435345858240023, &
-0.380878108274565, &
-0.262297552364679, &
-0.140138184037810, &
-0.0358748918526121, &
0.0488639928750531, &
0.123089717758417, &
0.185508107133491, &
0.228140113337704, &
0.243297948051216, &
0.228140113337704, &
0.18550810713349, &
0.123089717758417, &
0.048863992875052, &
-0.0358748918526117, &
-0.14013818403781, &
-0.262297552364679, &
-0.380878108274564, &
-0.435345858240023, &
-0.486008290350662, &
-0.534128406327251, &
-0.579323980561703, &
-0.622501288563042, &
-0.664795100033691, &
-0.706371081300489, &
-0.747708252680297, &
!CVDRIFT
-0.673036466796932, &
-0.631588446606282, &
-0.589736714506845, &
-0.546956675636816, &
-0.503115854789239, &
-0.457086814279026, &
-0.407967673072361, &
-0.356178260156541, &
-0.300454560238858, &
-0.179049058381449, &
-0.0536813324086593, &
0.053997808035211, &
0.142047789205761, &
0.219086341003571, &
0.283581066041592, &
0.327483912708784, &
0.343068685451976, &
0.327483912708784, &
0.283581066041591, &
0.219086341003571, &
0.14204778920576, &
0.0539978080352114, &
-0.0536813324086592, &
-0.179049058381449, &
-0.300454560238857, &
-0.356178260156542, &
-0.407967673072361, &
-0.457086814279024, &
-0.503115854789239, &
-0.546956675636816, &
-0.589736714506848, &
-0.631588446606287, &
-0.673036466796932, &
!GDS2
2.29008065199946, &
2.22942331101754, &
2.17329759898412, &
2.12243750433447, &
2.07575083944781, &
2.03066732241664, &
1.98427619082273, &
1.93382433897332, &
1.87440433662686, &
1.71594328315711, &
1.49687624300548, &
1.25252362054053, &
1.0182051580772, &
0.795171227114309, &
0.590999602469358, &
0.440107537273544, &
0.383997407012626, &
0.440107537273546, &
0.59099960246936, &
0.795171227114309, &
1.0182051580772, &
1.25252362054053, &
1.49687624300548, &
1.71594328315711, &
1.87440433662686, &
1.93382433897332, &
1.98427619082273, &
2.03066732241664, &
2.07575083944781, &
2.12243750433447, &
2.17329759898412, &
2.22942331101755, &
2.29008065199947, &
!GDS21
0.223719396038408, &
0.212584647182999, &
0.199986955006285, &
0.185460874109635, &
0.169330198130758, &
0.151352258125506, &
0.13122490348833, &
0.109157167850341, &
0.0846311406848634, &
0.028926919772887, &
-0.0335943556555319, &
-0.0967948099520018, &
-0.149551630676875, &
-0.172617273633045, &
-0.151761700834844, &
-0.0885667800428974, &
0.0, &
0.0885667800428985, &
0.151761700834845, &
0.172617273633045, &
0.149551630676875, &
0.0967948099520021, &
0.0335943556555319, &
-0.028926919772887, &
-0.0846311406848633, &
-0.109157167850341, &
-0.13122490348833, &
-0.151352258125505, &
-0.169330198130759, &
-0.185460874109636, &
-0.199986955006287, &
-0.212584647183001, &
-0.223719396038409, &
!Theta
-3.14159265358979, -2.96909691483081, -2.79753624140583, &
-2.62691063331481, -2.45995122460437, -2.29564390653795, &
-2.13398867911553, -1.97902030603011, -1.82790760022706, &
-1.54168052849789, -1.27627882252284, -1.03444601532502, &
-0.809104557779855, -0.596535696876655, -0.393953292215471, &
-0.19590929198706, -4.44089209850063e-16, 0.195909291987062, &
0.393953292215472, 0.596535696876654, 0.809104557779858, &
1.03444601532502, 1.27627882252284, 1.54168052849789, 1.82790760022705, &
1.97902030603012, 2.13398867911553, 2.29564390653794, 2.45995122460437, &
2.62691063331481, 2.79753624140584, 2.96909691483083, 3.14159265358979, &
!NULL
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000, &
0.0000000000000000  &
], (/33,10/))
end function miller_result

end program test_theta_grid
